unit Watch;
{
The TWatch Component V1.10

  MOST IMPORTANT :)
  =================
  This is NOT Freeware: It's PostCardWare. When you use
  this component or think it's useful, send me a post-card
  to (this is my parents' address, it will always reach me :)

  Florian Bmers
  Colmarer Str.11
  28211 Bremen
  GERMANY

  And of course, I am very interested in any application
  that uses this component (or any other application you wrote).
  If so, mail me (not the program, just an URL or similar) !
  (mail address below)

  Installation:
  =============
  1. Copy the files Watch.pas and watch.dcr to the directory where
     you store your components (or let it where it is)
  2. In Delphi, select Component|Install Component. In the
     following dialog enter the path and filename of
     Watch.pas and hit OK.
  3. Now the TWatch Component is available in the
     Component palette under Bome.
  4. Load watchdll.dpr into Delphi and compile it. It creates
     the dll "watchdll.dll".
  5. This library must be in the search path or in the directory of 
     the program in which you want to use TWatch. The dll is loaded 
     and unloaded dynamically when needed, in order to prevent Delphi
     to load it when installing the component.

  Description
  ===========
  This component enables you to track some system events.
  Currently implemented are Shell events (i.e. Activation,
  Creation, Closure of Applications), and Keyboard events
  (when something is typed in another application) and
  Mouse events (everything what the mouse does).

  The shell events have one drawback: The name parameter is
  identical for all instances of a watched program. You
  can only use the Handle parameter of the event to distinguish
  different instances of the same application.
  
  This component is designed for Delphi 2, 3, 4. It is not tested
  with Delphi 5.

  Note: the source for watchdll.dll is in the project watchdll.dpr.

  How to use it
  =============
  Be sure that only one instance/program uses this component
  at any time. You can use TOneInstance for that instance,
  you can download that from http://www.bome.com/ in the
  delphi components area.

  Different to previous versions of this component, you
  must now call the Start and Stop methods in order to
  start / stop watching of a specific type.

  Your main program MUST make sure that this component is
  freed properly. The easiest way of doing this is to put a
  line like "Watch1.Free;" in the main forms OnClose or OnDestroy
  event handler. Delphi seems to "forget" to free components'
  when closing the application.

  Properties:
  - NotifyOwnEvents: Whether an event shall be sent when
                     the own application receives a shell event
  - RepeatedEvents: When you activate different windows
                    of the same application, each time an
                    event is sent. If you set this to false,
                    only the first event will be sent.

  Events:
  - OnActivateShellWindow: Occurs when the shell window is activated
    (i.e. the desktop)
  - OnWindowActivated: Occurs when any program is activated
  - OnWindowCreated: Occurs when a program starts
  - OnWindowDestroyed: Occurs when a program closes
  - OnKeyNative: A raw key event, occurs everytime something with the
    keyboard happens.
  - OnMouseDown: Occurs when a mouse button is pressed
  - OnMouseMove: Occurs when the mouse moves
  - OnMouseUp: Occurs when a mouse button is released

  Note on Mouse events:
  Even if they contain the ShiftState, the Shift and Ctrl values will
  never be set, regardless whether these keys are actually pressed
  or not. You could combine it with a key hook to generate the "right"
  mouse events.

  ideas for enhancements:
  - make it usable for multiple applications...
  - of course, add the other hooks
  - use toAscii in order to get an Ascii-Character of a key
  - let the dll get some information about the window
    in which are typed the keys
  - possibility to change messages before they are processed by the
    watched application
  - let the dll change specific things while in context
    of another program e.g. add to every menu an own menu...
    or just get some information about that program

  Copyright
  =========
  (c) 1997-2000 by Florian Bmers

  send any comments, proposals, enhancements etc. to:
  delphi@bome.com

  new versions on:
  http://www.bome.com/

  file history (started 12/99):
  17 Dec 1999: - renamed watching.dll to watchdll.dll (virus scanners
                 detected it as a virus)
               - added mouse hook (thanks to David Hessler)
               - changed structure to use Start and Stop
               - minor changes in watchdll.dll
               - general clean up of code (it was 2 years old ;)
               - better support for Delphi 4 and Delphi 2
}
interface
uses SysUtils, Classes, Forms, Windows, Dialogs, Messages, Controls;

const
  watchdll='watchdll.dll';
  MSG_WATCH_SHELL_EVENT='WATCH_SHELL_EVENT';
  MSG_WATCH_KEY_EVENT='WATCH_KEY_EVENT';
  MSG_WATCH_MOUSE_EVENT='WATCH_MOUSE_EVENT';

type
  TWatchType=( wtSHELL,wtKEYBOARD, wtMouse, //implemented
               wtCALLWNDPROC,wtGETMESSAGE,  //not yet implemented
               wtMSGFILTER,wtSYSMSGFILTER); //not yet implemented

  TKeyInfo=record
   RepeatCount:Integer;
   ScanCode:Integer;
   Extended:Boolean;
   AltDown:Boolean;
   PreviousDown:Boolean;
   Down:Boolean;
   KeyNameText:String;
  end;

  //procedures in the dll
  TWatching=procedure(WatchType:TWatchType); stdcall;
  TStopAll=procedure; stdcall;
  TStillWatching=function:Boolean; stdcall;
  TSetReceiver=function(r:THandle):Boolean; stdcall;
  TGetReceiver=function:THandle; stdcall;

  // procedures for events
  TShellEvent=procedure( Sender:TObject;
                         Handle:Integer;
                         AppName:String) of Object;
  TKeyNativeEvent=procedure( Sender:TObject;
                             wParam,lParam:Integer) of Object;

  TWatch = class(TComponent)
  private
   // invisible window that receives messages from the dll
   HWNDHandle:THandle;
   // handle of the library
   LibHandle:THandle;
   DontLoadLib:Boolean;
   // message IDs for events of the dll
   SHELL_EVENT:DWORD;
   KEY_EVENT:DWORD;
   MOUSE_EVENT:DWORD;
   // procedures in the DLL
   StartWatching:TWatching;
   StopWatching:TWatching;
   StopAll:TStopAll;
   StillWatching:TStillWatching;
   SetReceiver:TSetReceiver;
   GetReceiver:TGetReceiver;

   {----- SHELL PROPERTIES ------}
   FNotifyOwnEvents:Boolean;
   FRepeatedEvents:Boolean;
   LastActivated:String;
   {----- SHELL EVENTS ------}
   FOnActivateShellWindow:TShellEvent;
   FOnWindowActivated: TShellEvent;
   FOnWindowCreated: TShellEvent;
   FOnWindowDestroyed: TShellEvent; //Window still lives when this call comes
   {----- KEY EVENTS --------}
   FOnKeyNative:TKeyNativeEvent;
   {----- Mouse EVENTS --------}
   FOnMouseDown:TMouseEvent;
   FOnMouseMove:TMouseMoveEvent;
   FOnMouseUp:TMouseEvent;

   procedure ResetLibVars;
   procedure LoadLib;
   procedure UnLoadLib;

  protected
   function GetAppName(window:THandle):String;
   procedure OnShellHook(nCode: Integer; wParam: Longint);
   procedure OnKeyHook(wParam,lParam: Longint);
   procedure OnMouseHook(wParam,lParam:Longint);

  public
   procedure OnHWNDEvent(var Message: TMessage);
   function getVKString(keyCode:Integer):String;
   procedure getKeyInfo(lParam:Integer; var KeyInfo:TKeyInfo);
   constructor Create(AOwner:TComponent); override;
   destructor Destroy; override;

   procedure Start(WatchType:TWatchType);
   procedure Stop(WatchType:TWatchType);
  published
   property NotifyOwnEvents:Boolean read FNotifyOwnEvents write FNotifyOwnEvents default false;
   property RepeatedEvents:Boolean read FRepeatedEvents write FRepeatedEvents default false;
   property OnActivateShellWindow: TShellEvent
            read FOnActivateShellWindow write FOnActivateShellWindow;
   property OnWindowActivated: TShellEvent
            read FOnWindowActivated write FOnWindowActivated;
   property OnWindowCreated: TShellEvent
            read FOnWindowCreated write FOnWindowCreated;
   property OnWindowDestroyed: TShellEvent
            read FOnWindowDestroyed write FOnWindowDestroyed;
   property OnKeyNative: TKeyNativeEvent read FOnKeyNative write FOnKeyNative;
   property OnMouseDown:TMouseEvent read FOnMouseDown write FOnMouseDown;
   property OnMouseMove:TMouseMoveEvent read FOnMouseMove write FOnMouseMove;
   property OnMouseUp:TMouseEvent read FOnMouseUp write FOnMouseUp;

  end;

procedure Register;

implementation

constructor TWatch.Create(AOwner:TComponent);
begin
 inherited Create(AOwner);
 // get unique Message IDs for the Broadcast of the dll
 SHELL_EVENT:=RegisterWindowMessage(PChar(MSG_WATCH_SHELL_EVENT));
 KEY_EVENT:=RegisterWindowMessage(PChar(MSG_WATCH_KEY_EVENT));
 MOUSE_EVENT:=RegisterWindowMessage(PChar(MSG_WATCH_MOUSE_EVENT));
end;

destructor TWatch.Destroy;
begin
 UnloadLib;
 inherited Destroy;
end;

procedure TWatch.UnloadLib;
begin
 try
  if assigned(GetReceiver) then
   if GetReceiver=HWNDHandle then
   begin
    SetReceiver(0);
    if assigned(StopAll) then
     StopAll;
   end;
  if HWNDHandle<>0 then
   DeallocateHwnd(HWNDHandle);
  if LibHandle<>0 then
   FreeLibrary(LibHandle);
 finally
  ResetLibVars;
 end;
end;

procedure TWatch.ResetLibVars;
begin
 LibHandle:=0;
 StartWatching:=nil;
 StopWatching:=nil;
 StopAll:=nil;
 StillWatching:=nil;
 SetReceiver:=nil;
 GetReceiver:=nil;
 HWNDHandle:=0;
end;

procedure TWatch.LoadLib;
begin
 if not dontLoadLib
    and (LibHandle=0)
    and  (not (csDesigning in ComponentState)) then
 begin
  dontLoadLib:=false;
  LibHandle:=LoadLibrary(watchdll);
  if LibHandle<>0 then
  begin
   // Create the window for notifications of the dll
   HWNDHandle:=AllocateHwnd(OnHWNDEvent);
   @StartWatching:=GetProcAddress(LibHandle, 'StartWatching');
   @StopWatching:=GetProcAddress(LibHandle, 'StopWatching');
   @StillWatching:=GetProcAddress(LibHandle, 'StillWatching');
   @StopAll:=GetProcAddress(LibHandle, 'StopAll');
   @SetReceiver:=GetProcAddress(LibHandle, 'SetReceiver');
   @GetReceiver:=GetProcAddress(LibHandle, 'GetReceiver');
   if assigned(SetReceiver) then
    if not SetReceiver(HWNDHandle) then
    begin
     MessageDlg('Another instance already uses '+watchdll+'.'+#13#10
      +'Overriding it.',mtError,[mbOK],0);
     SetReceiver(0);
     if not SetReceiver(HWNDHandle) then
     begin
      MessageDlg('Unable to override.'+#13#10
      +'System watching is disabled.',mtError,[mbOK],0);
      dontLoadLib:=true;
     end;
    end;
  end;
  if (LibHandle=0)
   or DontLoadLib
   or not assigned(StartWatching)
   or not assigned(StopWatching)
   or not assigned(StillWatching)
   or not assigned(StopAll)
   or not assigned(SetReceiver)
   or not assigned(GetReceiver) then
  begin
   if LibHandle=0 then
    MessageDlg('Library '+watchdll+' not found.'+#13#10
     +'System watching is disabled.',mtError,[mbOK],0)
   else
   if not dontLoadLib then
    MessageDlg('Library '+watchdll+' is not valid.'+#13#10
     +'System watching is disabled.',mtError,[mbOK],0);
   DontLoadLib:=true;
   UnloadLib;
  end;
 end;
end;

// helper method for the shell events
function TWatch.GetAppName(window:THandle):String;
//extra variable for clean conversion of PChar to String
var r:String;
begin
 setString(r,nil,300);
 if GetWindowText(window,PChar(r),300)=0 then
  result:=''
 else
  result:=PChar(r);
end;

procedure TWatch.OnShellHook(nCode: Integer; wParam: Longint);
var appName:String;
begin
 If NotifyOwnEvents or (Application.Handle<>DWORD(wParam)) then
 begin
  appName:=GetAppName(wParam);
  case nCode of
  HSHELL_ACTIVATESHELLWINDOW:
     begin
      if assigned(FOnActivateShellWindow) then
       FOnActivateShellWindow(Self,wParam,appName);
     end;
  HSHELL_WINDOWACTIVATED:
     begin
      if assigned(FOnWindowActivated) then
      begin
       if FRepeatedEvents or (appName<>LastActivated) then
        FOnWindowActivated(Self,wParam,appName);
       LastActivated:=appName;
      end;
     end;
  HSHELL_WINDOWCREATED:
     begin
      if assigned(FOnWindowCreated) then
       FOnWindowCreated(Self,wParam,appName);
     end;
  HSHELL_WINDOWDESTROYED:
     begin
      if assigned(FOnWindowDestroyed) then
       FOnWindowDestroyed(Self,wParam,appName);
     end;
  end;
 end;
end;

procedure TWatch.OnKeyHook(wParam,lParam: Longint);
begin
 if assigned(FOnKeyNative) then
  FOnKeyNative(Self,wParam,lParam);
end;

procedure TWatch.OnMouseHook(wParam,lParam: Longint);
var x,y:Integer;
    Button:TMouseButton;
    Shift:TShiftState;
    down:Boolean;
begin
 // the lParam contains the coordinates
 x:=lParam and $FFFF;
 y:=lParam shr 16;
 Button:=mbLeft;
 Shift:=[];
 down:=true;
 // I don't know where these constants are defined... too bad
 case (wParam and $F) of
 0: if assigned(FOnMouseMove) then
    begin
     FOnMouseMove(Self, [], x,y);
     exit;
    end;
 1: begin down:=true; Button:=mbLeft; end;
 2: begin down:=false; Button:=mbLeft; end;
 3: begin down:=true; Button:=mbLeft; Shift:=[ssDouble]; end;
 4: begin down:=true; Button:=mbRight; end;
 5: begin down:=false; Button:=mbRight; end;
 6: begin down:=true; Button:=mbRight; Shift:=[ssDouble]; end;
 7: begin down:=true; Button:=mbMiddle; end;
 8: begin down:=false; Button:=mbMiddle; end;
 9: begin down:=true; Button:=mbMiddle; Shift:=[ssDouble]; end;
 end;
 if down and assigned(FOnMouseDown) then
  FOnMouseDown(Self, Button, Shift, x, y)
 else
 if not down and assigned(FOnMouseUp) then
  FOnMouseUp(Self, Button, Shift, x, y);
end;

{This is the function to which the events of the dll are posted}
procedure TWatch.OnHWNDEvent(var Message: TMessage);
begin
 if Message.Msg=SHELL_EVENT then
  OnShellHook(Message.lParam,Message.wParam)
 else
 if Message.Msg=KEY_EVENT then
  OnKeyHook(Message.wParam,Message.lParam)
 else
 if Message.Msg=MOUSE_EVENT then
  OnMouseHook(Message.wParam,Message.lParam)
 else
 with Message do
  Result := DefWindowProc(HWNDHandle, Msg, wParam, lParam);
end;

procedure TWatch.Start(WatchType:TWatchType);
begin
 LoadLib;
 if assigned(StartWatching) then
  StartWatching(WatchType);
end;

procedure TWatch.Stop(WatchType:TWatchType);
begin
 if (Self.libHandle<>0)
   and assigned(StopWatching)
   and assigned(StillWatching) then
 begin
  StopWatching(WatchType);
  if not StillWatching then
   UnLoadLib;
 end;
end;

function TWatch.getVKString(keyCode:Integer):String;
var i:Integer;
begin
 result:='undefined 0x'+IntToHex(keyCode,4);
 i:=MapVirtualKey(keyCode,2);
 if (i<>0) and (i<$FF) then
  result:=chr(byte(i))
 else
 if ((keyCode>=$30) and (keyCode<=$39))
  or ((keyCode>=$41) and (keyCode<=$5A)) then
  result:=chr(byte(keyCode))
 else
 if ((keyCode>=$70) and (keyCode<=$87)) then
  result:='F'+IntToStr(keyCode-$6F)
 else
 if ((keyCode>=$60) and (keyCode<=$69)) then
  result:='Numeric keypad '+IntToStr(keyCode-$60)
 else
 case keyCode of
  VK_BACK: result:='BACKSPACE';
  VK_TAB:  result:='TAB';
  VK_CLEAR: result:='CLEAR';
  VK_RETURN: result:='ENTER';
  VK_SHIFT: result:='SHIFT';
  VK_CONTROL: result:='CTRL';
  VK_MENU: result:='ALT';
  VK_PAUSE: result:='PAUSE';
  VK_CAPITAL: result:='CAPS LOCK';
  VK_ESCAPE: result:='ESC';
  VK_SPACE: result:='SPACEBAR';
  VK_PRIOR: result:='PAGE UP';
  VK_NEXT: result:='PAGE DOWN';
  VK_END: result:='END';
  VK_HOME: result:='HOME';
  VK_LEFT: result:='LEFT ARROW';
  VK_UP: result:='UP ARROW';
  VK_RIGHT: result:='RIGHT ARROW';
  VK_DOWN: result:='DOWN ARROW';
  VK_SELECT: result:='SELECT';
  VK_EXECUTE: result:='EXECUTE';
  VK_SNAPSHOT: result:='PRINT SCREEN';
  VK_INSERT: result:='INS';
  VK_DELETE: result:='DEL';
  VK_HELP: result:='HELP';
  VK_MULTIPLY: result:='MULTIPLY';
  VK_ADD: result:='ADD';
  VK_SEPARATOR: result:='SEPARATOR';
  VK_SUBTRACT: result:='SUBTRACT';
  VK_DECIMAL: result:='DECIMAL';
  VK_DIVIDE: result:='DIVIDE';
  VK_NUMLOCK: result:='NUM LOCK';
  VK_SCROLL: result:='SCROLL LOCK';
 end;
end;


procedure TWatch.getKeyInfo(lParam:Integer; var KeyInfo:TKeyInfo);
var s:String;
begin
 s:='';
 setString(s,nil,80);
 GetKeyNameText(lParam,PChar(S),79);
 setlength(s,strlen(PChar(s)));
 with KeyInfo do
 begin
  RepeatCount:=lParam and $FFFF;
  ScanCode:=(lParam and $FF0000) shr 16;
  Extended:=((lParam shr 24) and 1)=1;
  AltDown:=((lParam shr 29) and 1)=1;
  PreviousDown:=((lParam shr 30) and 1)=1;
  Down:=((lParam shr 31) and 1)=0;
  KeyNameText:=s;
 end;
end;

{procedure TWatch.SetDLLReceiver(r:Integer);
begin
 if assigned(SetReceiver) and assigned(GetReceiver) then
 begin

  SetReceiver(r);
end;}

procedure Register;
begin
 RegisterComponents('Bome',[TWatch]);
end;

end.
